@@ -1,10 +1,13 @@
+import warnings
 import maui.utils
 import numpy as np
 import pandas as pd
 from functools import partial
+from keras import backend as K
 from scipy import spatial, cluster
 from sklearn.cluster import KMeans
 from sklearn.base import BaseEstimator
+from .maui_warnings import MauiWarning
 from .autoencoders_architectures import stacked_vae, deep_vae, train_model
 
 
@@ -34,6 +37,10 @@ class Maui(BaseEstimator):
         the intermediate layers are also variational. If 'deep', will train a deep VAE
         where the intermediate layers are regular (ReLU) units, and only the middle
         (latent) layer is variational.
+
+    verbose:
+         Integer. 0, 1, or 2. Verbosity mode. 0 = silent, 1 = progress bar, 2 = one line per epoch.
+         default: 0
     """
 
     def __init__(
@@ -54,6 +61,7 @@ class Maui(BaseEstimator):
         relu_intermediaries=True,
         relu_embedding=True,
         input_dim=None,
+        verbose=0,
     ):
         if n_hidden is None:
             n_hidden = [1500]
@@ -73,6 +81,7 @@ class Maui(BaseEstimator):
                 epsilon_std=epsilon_std,
                 batch_normalize_inputs=batch_normalize_inputs,
                 batch_normalize_intermediaries=batch_normalize_intermediaries,
+                batch_normalize_embedding=batch_normalize_embedding,
                 relu_intermediaries=relu_intermediaries,
                 relu_embedding=relu_embedding,
             )
@@ -113,9 +122,12 @@ class Maui(BaseEstimator):
             batch_size=batch_size,
             kappa=kappa,
             max_beta_val=max_beta_val,
+            verbose=verbose,
         )
 
-    def fit(self, X, y=None, X_validation=None):
+        self.feature_names = None
+
+    def fit(self, X, y=None, X_validation=None, *args, **kwargs):
         """Train autoencoder model
 
         Parameters
@@ -132,9 +144,17 @@ class Maui(BaseEstimator):
         -------
         self : Maui object
         """
-        self.x_ = self._dict2array(X)
+        x_ = self._dict2array(X)
+        self._validate_indices(x_)
+        self.x_ = x_
+
+        self.nwp_ = self.feature_correlations_ = self.w_ = None
+
         x_test = self._dict2array(X_validation) if X_validation else self.x_
 
+        if self.feature_names != x_test.columns.tolist():
+            raise ValueError("Feature mismatch between X and X_val!")
+
         if not hasattr(self, "vae"):
             vae, encoder, sampling_encoder, decoder, beta = self.architecture(
                 self.x_.shape[1],
@@ -149,7 +169,9 @@ class Maui(BaseEstimator):
             self.encoder = encoder
             self.sampling_encoder = sampling_encoder
             self.decoder = decoder
-        hist = self.training_fn(vae=vae, x_train=self.x_, x_val=x_test, beta=beta)
+        hist = self.training_fn(
+            vae=self.vae, x_train=self.x_, x_val=x_test, beta=self.beta, **kwargs
+        )
         self.hist = pd.DataFrame(hist.history)
         return self
 
@@ -180,15 +202,16 @@ class Maui(BaseEstimator):
         else:
             raise ValueError("`encoder` must be one of 'mean' or 'sample'")
 
-        self.x_ = self._dict2array(X)
+        x_ = self._dict2array(X)
+        self._validate_indices(x_)
+        self.x_ = x_
         self.z_ = pd.DataFrame(
             the_encoder.predict(self.x_),
             index=self.x_.index,
             columns=[f"LF{i}" for i in range(1, self.n_latent + 1)],
         )
-        self.feature_correlations = maui.utils.correlate_factors_and_features(
-            self.z_, self.x_
-        )
+
+        self.feature_correlations_ = None
         self.w_ = None
         return self.z_
 
@@ -214,6 +237,62 @@ class Maui(BaseEstimator):
         self.fit(X, X_validation=X_validation, y=y)
         return self.transform(X, encoder=encoder)
 
+    def fine_tune(
+        self,
+        X,
+        batch_size=100,
+        epochs=100,
+        initial_beta_val=1,
+        kappa=1.0,
+        max_beta_val=1,
+        learning_rate=0.0005,
+        verbose=0,
+        **kwargs,
+    ):
+        """Fine-tune autoencoder model. This is meant to be used with a pre-trained model.
+
+        Parameters
+        ----------
+        X:      dict with multi-modal dataframes, containing training data, e.g.
+                    {'mRNA': df1, 'SNP': df2},
+                    df1, df2, etc. are (n_features, n_samples) pandas.DataFrame's.
+                    The sample names must match, the feature names need not.
+
+        batch_size: int (default 100)
+                    The size of the mini-batches used for training the network
+
+        epochs:     int (default 100)
+                    The number of epoches to use for training the network
+
+        Returns
+        -------
+        self : Maui object
+        """
+        if not hasattr(self, "vae"):
+            raise Exception(
+                "Cannot call fine_tune() on a model that hasn't been trained yet!"
+            )
+        x_ = self._dict2array(X)
+        self._validate_indices(x_)
+        self.x_ = x_
+
+        self.nwp_ = self.feature_correlations_ = self.w_ = None
+
+        fine_tunine_fn = partial(
+            train_model,
+            epochs=epochs,
+            batch_size=batch_size,
+            kappa=kappa,
+            max_beta_val=max_beta_val,
+            verbose=verbose,
+        )
+        K.set_value(self.beta, initial_beta_val)
+        hist = fine_tunine_fn(
+            vae=self.vae, x_train=self.x_, x_val=self.x_, beta=self.beta, **kwargs
+        )
+        self.hist = pd.DataFrame(hist.history)
+        return self
+
     def cluster(
         self,
         k=None,
@@ -287,8 +366,9 @@ class Maui(BaseEstimator):
                 name=score_name,
             )
             self.kmeans_scores.index.name = "K"
-            self.optimal_k_ = np.argmax(self.kmeans_scores)
-            self.yhat_ = yhats[self.optimal_k_]
+            opt_k_index = np.argmax(self.kmeans_scores)
+            self.optimal_k_ = self.kmeans_scores.index[opt_k_index]
+            self.yhat_ = yhats[opt_k_index]
             return self.yhat_
 
     def compute_roc(self, y, **kwargs):
@@ -447,6 +527,39 @@ class Maui(BaseEstimator):
             )
         return self.w_
 
+    def get_feature_correlations(self):
+        """Get correlation coefficients between input features and latent factors.
+
+        Returns
+        -------
+        r:  (n_features, n_latent_factors) DataFrame
+            r_{ij} is the correlation coefficient between feature `i`
+            and latent factor `j`.
+        """
+        if (
+            not hasattr(self, "feature_correlations_")
+            or self.feature_correlations_ is None
+        ):
+            self.feature_correlations_ = maui.utils.correlate_factors_and_features(
+                self.z_, self.x_
+            )
+        return self.feature_correlations_
+
+    def get_neural_weight_product(self):
+        """Get the product of neural weights from feature input to latent factors.
+
+        Returns
+        -------
+        nwp:    (n_features, n_latent_factors) DataFrame
+                nwp_{ij} is the product of the weights in the encoder network
+                along the path leading from input feature `i` to latent
+                factor `j`.
+
+        """
+        if not hasattr(self, "nwp_") or self.nwp_ is None:
+            self.nwp_ = maui.utils.neural_path_weight_product(self)
+        return self.nwp_
+
     def drop_unexplanatory_factors(self, threshold=0.02):
         """Drops factors which have a low R^2 score in a univariate linear model
         predicting the features `x` from a column of the latent factors `z`.
@@ -552,9 +665,13 @@ class Maui(BaseEstimator):
         with open(os.path.join(destdir, "maui_args.json"), "wt") as outfile:
             json.dump(self.init_args, outfile)
         self.vae.save_weights(os.path.join(destdir, "maui_weights.h5"))
+        with open(os.path.join(destdir, "maui_feature_names.txt"), "wt") as outfile:
+            for fn in self.feature_names:
+                outfile.write(fn)
+                outfile.write("\n")
 
-    @staticmethod
-    def load(directory):
+    @classmethod
+    def load(cls, directory):
         """Load a maui model from disk, which was previously saved using ``save()``
 
         Parameters
@@ -572,8 +689,19 @@ class Maui(BaseEstimator):
 
         with open(os.path.join(directory, "maui_args.json"), "rt") as infile:
             init_args = json.load(infile)
-        maui_model = Maui(**init_args)
+        maui_model = cls(**init_args)
         maui_model.vae.load_weights(os.path.join(directory, "maui_weights.h5"))
+        if os.path.isfile(os.path.join(directory, "maui_feature_names.txt")):
+            with open(
+                os.path.join(directory, "maui_feature_names.txt"), "rt"
+            ) as infile:
+                maui_model.feature_names = [l.strip() for l in infile]
+        else:
+            warnings.warn(
+                MauiWarning(
+                    "Feature names not found in save dir. Loaded model will not perform feature name validation on first call!"
+                )
+            )
         return maui_model
 
     def _validate_X(self, X):
@@ -589,15 +717,24 @@ class Maui(BaseEstimator):
 
         return True
 
-    def _validate_indices(self):
-        pass
+    def _validate_indices(self, x):
+        if self.feature_names is None:
+            self.feature_names = x.columns.tolist()
+        else:
+            if self.feature_names != x.columns.tolist():
+                raise ValueError("Feature mismatch with previously fit data!")
 
     def _dict2array(self, X):
         self._validate_X(X)
         new_feature_names = [f"{k}: {c}" for k in sorted(X.keys()) for c in X[k].index]
         sample_names = X[list(X.keys())[0]].columns
-        return pd.DataFrame(
+        ret = pd.DataFrame(
             np.vstack([X[k] for k in sorted(X.keys())]).T,
             index=sample_names,
             columns=new_feature_names,
         )
+        if self.feature_names is not None:
+            if set(self.feature_names) != set(ret.columns):
+                raise ValueError("Feature mismatch!")
+            ret = ret.loc[:, self.feature_names]
+        return ret
